@@ -43,6 +43,7 @@ class IClient(Generic[Request, Response]):
     request_class: type[IRequest[Request]]
     response_class: type[IResponse[Request, Response]]
     logger: logging.Logger = logging.getLogger('headless.client')
+    user_agent: str = 'Headless'
 
     def check_json(self, headers: Headers):
         # TODO: Abstract this to a separate class.
@@ -53,6 +54,18 @@ class IClient(Generic[Request, Response]):
                 f'{headers.get("Content-Type")}'
             )
 
+    def in_context(self) -> bool:
+        """Return a boolean indicating if the client is current used as a
+        context manager.
+        """
+        raise NotImplementedError
+
+    async def get(self, **kwargs: Any) -> IResponse[Request, Response]:
+        return await self.request(method='GET', **kwargs)
+
+    async def post(self, **kwargs: Any) -> IResponse[Request, Response]:
+        return await self.request(method='POST', **kwargs)
+
     async def persist(
         self,
         model: type[M],
@@ -71,16 +84,20 @@ class IClient(Generic[Request, Response]):
         method: str,
         url: str,
         credential: ICredential | None = None,
-        json: list[Any] | dict[str, Any] | None = None
+        json: list[Any] | dict[str, Any] | None = None,
+        params: dict[str, Any] | None = None,
+        headers: dict[str, str] | None = None
     ) -> IResponse[Request, Response]:
-        headers: dict[str, str] = {}
+        headers: dict[str, str] = headers or {}
+        headers.setdefault('User-Agent', self.user_agent)
         if json is not None:
             headers['Content-Type'] = 'application/json'
         request = await self._request_factory(
             method=method,
             url=url,
             headers=headers,
-            json=json
+            json=json,
+            params=params
         )
         await (credential or self.credential).add_to_request(request)
         try:
@@ -93,13 +110,14 @@ class IClient(Generic[Request, Response]):
             response = await self.on_rate_limited(response)
         return response
 
-    async def retrieve(self, model: type[M], resource_id: int | str) -> M:
+    async def retrieve(self, model: type[M], resource_id: int | str | None = None) -> M:
         """Discover the API endpoint using the class configuration
         and retrieve a single instance using the HTTP GET verb.
         """
-        response = await self.request(
-            method='GET',
-            url=model._meta.get_retrieve_url(resource_id) # type: ignore
+        meta = model.get_meta()
+        response = await self.get(
+            url=model.get_retrieve_url(resource_id),
+            headers=meta.headers
         )
         response.raise_for_status()
         self.check_json(response.headers)
@@ -123,11 +141,15 @@ class IClient(Generic[Request, Response]):
             raise RateLimited(request=response.request, response=response)
         return response
 
-    def process_response(self, action: str, data: dict[str, Any] | list[Any]) -> dict[str, Any]:
+    def process_response(
+        self,
+        action: str,
+        data: dict[str, Any] | list[Any]
+    ) -> dict[str, Any]:
         """Hook to transform response data."""
         return data
 
-    def resource_factory(self, model: type[M], action, data: dict[str, Any]) -> M:
+    def resource_factory(self, model: type[M], action: str, data: dict[str, Any]) -> M:
         resource = model.parse_obj(model.process_response(action, data))
         self._inject_client(resource)
         return resource
@@ -137,7 +159,8 @@ class IClient(Generic[Request, Response]):
         method: str,
         url: str,
         json: list[Any] | dict[str, Any] | None = None,
-        headers: dict[str, str] | None = None
+        headers: dict[str, str] | None = None,
+        params: dict[str, Any] | None = None
     ) -> Request:
         raise NotImplementedError
 
@@ -147,7 +170,7 @@ class IClient(Generic[Request, Response]):
     async def __aenter__(self: T) -> T:
         raise NotImplementedError
 
-    async def __aexit__(self, cls: type[BaseException], *args: Any) -> bool | None:
+    async def __aexit__(self, cls: type[BaseException] | None, *args: Any) -> bool | None:
         raise NotImplementedError
 
     def _inject_client(self, resource: pydantic.BaseModel):
@@ -166,7 +189,10 @@ class IClient(Generic[Request, Response]):
                     self._inject_client(subresource)
 
     async def _request_factory(self, *args: Any, **kwargs: Any) -> IRequest[Request]:
-        request = await self.request_factory(*args, **kwargs)
+        request = await self.request_factory(
+            *args,
+            **(await self.credential.preprocess_request(**kwargs))
+        )
         return self.request_class.fromimpl(request)
 
     async def listall(
@@ -176,9 +202,11 @@ class IClient(Generic[Request, Response]):
         url: str | None = None
     ) -> Generator[M, None, None]:
         """Like :meth:`list()`, but returns all entities."""
+        meta = model.get_meta()
         response = await self.request(
             method='GET',
-            url=url or model.get_list_url(*params)
+            url=url or model.get_list_url(*params),
+            headers=meta.headers
         )
         response.raise_for_status()
         self.check_json(response.headers)
@@ -200,13 +228,16 @@ class IClient(Generic[Request, Response]):
         """Discover the API endpoint using the class configuration
         and retrieve a list of instances using the HTTP GET verb.
         """
+        meta = model.get_meta()
         response = await self.request(
             method='GET',
-            url=model._meta.get_list_url()
+            url=model.get_list_url(),
+            headers=meta.headers
         )
         response.raise_for_status()
         self.check_json(response.headers)
         data = self.process_response('list', await response.json())
+        if inspect.isawaitable(data): data = await data
         data = model.process_response('list', data)
         resources = [
             self.resource_factory(model, None, x)
